import json
from gi.repository import Gio

from . import getLogger

logger = getLogger(__name__)


class MrgCommand():
    def init_command(self, command_socket=None):

        self.__cancellable = None
        self.__gsocket = None
        self.__conn = None
        self.__data_input = None

        # List of outgoing commands
        self.__commands = []

        # Current outgoing command buffer
        self.__outgoing_data = None

        # Incoming command message length
        self.__incoming_total = 0

        # Incoming buffer
        self.__incoming_data = None

        if command_socket:
            self.init_connection(command_socket)

    def __on_data_output(self, source, res):
        try:
            status, bytes_written = source.write_all_finish(res)
        except Exception as e:
            if e.code != Gio.IOErrorEnum.CANCELLED:
                logger.warning(f"Error writing command {e}")
            self.close_connection()
            return

        self.__outgoing_data = None
        self.__consume_next_command()

    def __consume_next_command(self):
        if self.__conn is None or self.__outgoing_data or not self.__commands:
            return

        command = self.__commands.pop(0)
        command_len = len(command)

        # Create command buffer by appending size to the beginning
        self.__outgoing_data = command_len.to_bytes(4) + command

        self.__conn.props.output_stream.write_all_async(
            self.__outgoing_data,
            0,
            self.__cancellable,
            self.__on_data_output
        )

    def write_command(self, command, args=None):
        cmd = {"command": command}

        if args is not None:
            cmd["args"] = args

        cmd = json.dumps(cmd)

        self.__commands.append(cmd.encode())
        self.__consume_next_command()

    def handle_command(self, line):
        pass

    def __on_data_input(self, source, res):
        try:
            data = source.read_bytes_finish(res)
        except Exception as e:
            if e.code != Gio.IOErrorEnum.CANCELLED:
                logger.warning(f"Error reading command {e}")
            self.close_connection()
            return

        if self.__incoming_data:
            self.__incoming_data = self.__incoming_data + data.get_data()
        else:
            self.__incoming_data = data.get_data()

        # Check if we are done reading data for command
        if data.get_size() != self.__incoming_total:
            self.__incoming_total -= data.get_size()

            # Read the rest
            source.read_bytes_async(self.__incoming_total, 0, self.__cancellable, self.__on_data_input)
            return

        if self.__incoming_data is not None:
            self.handle_command(self.__incoming_data.decode("UTF-8"))
            self.__incoming_total = 0
            self.__incoming_data = None

        # Queue next command read
        source.read_bytes_async(4, 0, self.__cancellable, self.__on_data_len_input)

    def __on_data_len_input(self, source, res):
        try:
            data = source.read_bytes_finish(res)
        except Exception as e:
            if e.code != Gio.IOErrorEnum.CANCELLED:
                logger.warning(f"Error reading command {e}")
            self.close_connection()
            return

        # Just in case
        if data.get_size() < 4:
            missing_data = source.read_bytes(4 - data.get_size(), None)
            self.__incoming_total = int.from_bytes(data.get_data() + missing_data.get_data())
        else:
            self.__incoming_total = int.from_bytes(data.get_data())

        # Read command data
        source.read_bytes_async(self.__incoming_total, 0, self.__cancellable, self.__on_data_input)

    def init_connection(self, fd):
        self.__cancellable = Gio.Cancellable()

        self.__gsocket = Gio.Socket.new_from_fd(fd)
        self.__conn = Gio.SocketConnection(socket=self.__gsocket)
        self.__data_input = Gio.DataInputStream.new(self.__conn.props.input_stream)

        # Queue first command read
        self.__data_input.read_bytes_async(4, 0, self.__cancellable, self.__on_data_len_input)

    def close_connection(self):
        # Cancel all pending IO
        if self.__cancellable:
            self.__cancellable.cancel()
            self.__cancellable = None

        if self.__conn:
            self.__conn.close()
            self.__conn = None
            self.__data_input = None
            self.__gsocket = None

        self.__commands = []
        self.__outgoing_data = None
        self.__incoming_total = 0
        self.__incoming_data = None



# GtkPaned Controller
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_widget import MrgGtkWidget
from merengue import MrgPlaceholder


class MrgGtkPaned(MrgGtkWidget):
    object = GObject.Property(type=Gtk.Paned, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def set_start_child(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 3:
            self.object.add1(child)
        else:
            self.object.set_start_child(child)

    def set_end_child(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 3:
            self.object.add2(child)
        else:
            self.object.set_end_child(child)

    def get_start_child(self):
        if self.object is None:
            return None

        if Gtk.MAJOR_VERSION == 3:
            return self.object.get_child1()
        else:
            return self.object.get_start_child()

    def get_end_child(self):
        if self.object is None:
            return None

        if Gtk.MAJOR_VERSION == 3:
            return self.object.get_child2()
        else:
            return self.object.get_end_child()

    def get_child_position(self, child):
        return 0 if child == self.get_start_child() else 1

    def __update_placeholder(self):
        if self.object is None:
            return

        start_child = self.get_start_child()
        end_child = self.get_end_child()

        if start_child is None:
            self.set_start_child(MrgPlaceholder(visible=True, controller=self))

        if end_child is None:
            self.set_end_child(MrgPlaceholder(visible=True, controller=self))

    def object_changed(self, old, new):
        super().object_changed(old, new)
        self.__update_placeholder()



# Merengue Adwaita plugin
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import gi

gi.require_version("Adw", "1")
from gi.repository import GObject, Adw

Adw.init()

from .mrg_adw_application_window import MrgAdwApplicationWindow
from .mrg_adw_bin import MrgAdwBin
from .mrg_adw_carousel import MrgAdwCarousel
from .mrg_adw_window import MrgAdwWindow
from .mrg_adw_dialog import MrgAdwDialog
from .mrg_adw_view_stack import MrgAdwViewStack
from .mrg_adw_view_stack_page import MrgAdwViewStackPage




# Cambalache Application
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import gi

gi.require_version('GIRepository', '3.0')

# flake8: noqa: E402,F401
from cambalache import config
from gi.repository import Gio

resource = Gio.Resource.load(os.path.join(config.pkgdatadir, "app.gresource"))
resource._register()

repository = gi.Repository.get_default()
repository.prepend_search_path(config.privatecambalachedir)
repository.prepend_library_path(config.privatecambalachedir)

gi.require_version('CambalachePrivate', '4.0')

from .cmb_application import CmbApplication
from .cmb_scrolled_window import CmbScrolledWindow
from .cmb_picture import CmbPicture
